require 'minitest/unit'
require 'minitest/spec'
require 'minitest/mock'
require 'minitest/autorun'
require 'fileutils'
require 'tempfile'
require_relative 'dorkbox'


class TestCreation < MiniTest::Unit::TestCase
  include Dorkbox

  def test_create
    Dir.mktmpdir() { |remote_repo_dir|
      `git init --bare #{remote_repo_dir}`

      Dir.mktmpdir() { |local_dorkbox_repo_dir|
        Dir.chdir(local_dorkbox_repo_dir) {
          create(remote_repo_dir)
          assert(File.exists?(GITIGNORE))
        }
      }

    }
    cleanup_tracked()
  end

  def test_connect
    Dir.mktmpdir() { |remote_repo_dir|
      `git init --bare #{remote_repo_dir}`

      first_client_name = nil
      Dir.mktmpdir() { |first_repo_dir|
        Dir.chdir(first_repo_dir) {
          first_client_name = create(remote_repo_dir)
        }

        Dir.mktmpdir() { |second_repo_dir|
          Dir.chdir(second_repo_dir) {
            second_client_name = connect(remote_repo_dir)
            all_branches = c("git branch -a")
            assert(all_branches.include?(first_client_name))
            assert(all_branches.include?(second_client_name))
          }
        }
      }
    }
    cleanup_tracked()
  end
end

class TestSync < MiniTest::Unit::TestCase
  include Dorkbox

  def setup
    @remote_repo_dir = Dir.mktmpdir()
    `git init --bare #{@remote_repo_dir}`

    @first_client_dir = Dir.mktmpdir()
    @second_client_dir = Dir.mktmpdir()

    Dir.chdir(@first_client_dir) {
      create(@remote_repo_dir)
    }

    Dir.chdir(@second_client_dir) {
      connect(@remote_repo_dir)
    }
  end

  def teardown
    FileUtils.remove_entry_secure(@remote_repo_dir)
    FileUtils.remove_entry_secure(@first_client_dir)
    FileUtils.remove_entry_secure(@second_client_dir)
    cleanup_tracked()
  end

  def test_syncing_between_two_clients
    Dir.chdir(@first_client_dir) {
      File.open("something", "w") { |f| f.write("asd") }
      sync()
    }

    Dir.chdir(@second_client_dir) {
      sync()
      File.open("something") { |f| assert_equal("asd", f.read()) }
    }

    Dir.chdir(@second_client_dir) {
      File.open("something", "a") { |f| f.write("xyz") }
      sync()
    }

    Dir.chdir(@first_client_dir) {
      sync()
      File.open("something", "r") { |f| assert_equal("asdxyz", f.read()) }
    }
  end

  def test_tracking_between_two_clients
    Dir.chdir(@first_client_dir) {
      File.open("something", "w") { |f| f.write("asd") }
      sync_tracked()
    }

    Dir.chdir(@second_client_dir) {
      sync_tracked()
      File.open("something") { |f| assert_equal("asd", f.read()) }
    }

    Dir.chdir(@second_client_dir) {
      File.open("something", "a") { |f| f.write("xyz") }
      sync_tracked()
    }

    Dir.chdir(@first_client_dir) {
      sync_tracked()
      File.open("something", "r") { |f| assert_equal("asdxyz", f.read()) }
    }
  end

  def test_conflicted_client_doesnt_sync_until_fixed
    Dir.chdir(@first_client_dir) {
      File.open("something", "w") { |f| f.write("asd") }
      sync()
    }

    Dir.chdir(@second_client_dir) {
      File.open("something", "w") { |f| f.write("whatsup") }
      begin
        sync()
      rescue StandardError => e
      end

      assert(File.exists? CONFLICT_STRING)

    }

    Dir.chdir(@second_client_dir) {
      begin
        sync()
        flunk('should have raised exception')
      rescue StandardError => e
      end
    }

    Dir.chdir(@second_client_dir) {
      `git merge dorkbox/master || true`
      File.open("something", "w") { |f| f.write("merged") }
      `git commit  -am 'merged'`
      File.delete(CONFLICT_STRING)
      sync()
    }

    Dir.chdir(@first_client_dir) {
      sync()
      File.open("something") { |f| assert_equal("merged", f.read()) }
    }
  end

  def test_multiple_sync_without_changes_doesnt_crash
    Dir.chdir(@second_client_dir) {
      sync()
      sync()
      sync()
    }

  end


end


class TestCrontabAdding < MiniTest::Unit::TestCase
  include Dorkbox

  def setup
    begin
      @save_user_crontab = c('crontab -l 2>/dev/null')
    rescue
      @save_user_crontab = nil
    end
    `crontab -r || /bin/true`
  end

  def teardown
    if @save_user_crontab.nil?
      `crontab -r`
    else
      tmp = Tempfile.new('dorkbox-temp')
      tmp.puts(@save_user_crontab)
      tmp.flush()
      `crontab #{tmp.path}`
      tmp.close()
    end
  end

  def test_dorkbox_cron_is_enabled_when_crontab_empty
    enable_dorkbox_cronjob()
    v = c('crontab -l')
    assert(v.scan(/#{DORKBOX_CRONTAB_COMMENT}/).size == 2)
  end

  def test_dorkbox_cron_is_not_duplicated_if_already_there
    enable_dorkbox_cronjob()
    enable_dorkbox_cronjob('asdasd')
    v = c('crontab -l')
    assert(v.scan(/#{DORKBOX_CRONTAB_COMMENT}/).size == 2)
  end

  def test_dorkbox_cron_is_updated_if_already_there
    enable_dorkbox_cronjob()
    enable_dorkbox_cronjob('asdasd')
    v = c('crontab -l')
    assert(v.scan(/#{DORKBOX_CRONTAB_COMMENT}/).size == 2)
    assert(v.scan(/asdasd/).size == 1)
  end
end


